"""Functionality to estimate performance of a strategy."""

from __future__ import annotations

__all__ = [
    "calculate",
    "to_returns",
]

import numpy as np
import pandas as pd

from pqr.utils import align


def calculate(
        holdings: pd.DataFrame,
        *,
        universe_returns: pd.DataFrame,
) -> pd.Series:
    """Calculates portfolio returns.

    Parameters
    ----------
    holdings : pd.DataFrame
        Weights of a portfolio.
    universe_returns : pd.DataFrame
        Returns of universe, available to trade for a strategy.

    Returns
    -------
    pd.Series
        Periodic returns of a portfolio.
    """

    holdings, universe_returns = align(holdings, universe_returns)
    returns = np.zeros(len(holdings))
    returns[1:] = np.asarray(holdings)[:-1] * np.asarray(universe_returns)[1:]
    return pd.Series(returns, index=holdings.index.copy())


def to_returns(prices: pd.DataFrame) -> pd.DataFrame:
    """Calculates universe returns from given prices of assets universe.

    1st period returns are set to zero. All incorrect values (nans and infs) are converted to zero.

    Parameters
    ----------
    prices : pd.DataFrame
        Matrix with **close** prices of traded assets and nans for missing assets (e.g. already
        delisted)

    Returns
    -------
    pd.DataFrame
        Returns of assets universe.
    """

    prices_array = np.asarray(prices)
    universe_returns = np.zeros_like(prices_array, dtype=float)
    universe_returns[1:] = np.diff(prices_array, axis=0) / prices_array[:-1]
    universe_returns = np.nan_to_num(universe_returns, nan=0, neginf=0, posinf=0)
    return pd.DataFrame(
        universe_returns,
        index=prices.index.copy(),
        columns=prices.columns.copy(),
    )



"""Implemented classic factor strategies and pretransforms for them."""

from __future__ import annotations

__all__ = [
    "filter",
    "look_back",
    "lag",
    "hold",
    "quantiles",
    "top",
    "bottom",
    "thresholds",
]

from typing import Literal, Callable

import numpy as np
import pandas as pd

from pqr.utils import align


def filter(
        factor: pd.DataFrame,
        *,
        universe: pd.DataFrame,
) -> pd.DataFrame:
    """Filters `factor` values based on given `universe`.

    Actually, replaces factor values with nans, where universe is equal to False.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    universe : pd.DataFrame
        Matrix with True/False, indicating to include factor values or not.

    Returns
    -------
    pd.DataFrame
        Filtered `factor` values.
    """

    universe, factor = align(universe, factor)
    return pd.DataFrame(
        np.where(np.asarray(universe, dtype=bool), np.asarray(factor, bool), np.nan),
        index=factor.index.copy(),
        columns=factor.columns.copy(),
    )


def look_back(
        factor: pd.DataFrame,
        *,
        period: int,
        agg: Literal["pct", "mean", "median", "min", "max"] | Callable[[pd.Series], float],
) -> pd.DataFrame:
    """Aggregates `factor` values column-wise each `period` by `agg`.

    If agg is not predefined can work very slow. In this case the best decision is to write your
    own effective realisation of transformation function.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    period : int
        Period to look back on the data.
    agg : {"pct", "mean", "median", "min", "max"} or callable
        Aggregation func to apply on `factor` values. If callable is given function must be
        appliable on pd.Series and return float.

    Returns
    -------
    pd.DataFrame
        Aggregated `factor` values.
    """

    if agg == "pct":
        factor_array = np.asarray(factor)
        abs_change = (factor_array[period:] - factor_array[:-period])
        base = factor_array[:-period]
        return pd.DataFrame(
            abs_change / base,
            index=factor.index[period:].copy(),
            columns=factor.columns.copy()
        )
    elif agg == "mean":
        return factor.rolling(period, axis=0).mean().iloc[period:]
    elif agg == "median":
        return factor.rolling(period, axis=0).median().iloc[period:]
    elif agg == "min":
        return factor.rolling(period, axis=0).min().iloc[period:]
    elif agg == "max":
        return factor.rolling(period, axis=0).max().iloc[period:]
    else:
        return factor.rolling(period, axis=0).apply(agg).iloc[period:]


def lag(
        factor: pd.DataFrame,
        *,
        period: int,
) -> pd.DataFrame:
    """Lags `factor` values for `period`.

    Can be used both forward and backward.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    period : int
        Period to look back on the data.

    Returns
    -------
    pd.DataFrame
        Lagged `factor` values.
    """

    if period == 0:
        return factor
    elif period > 0:
        return pd.DataFrame(
            factor.to_numpy()[:-period],
            index=factor.index[period:].copy(),
            columns=factor.columns.copy(),
        )
    else:  # period < 0
        return pd.DataFrame(
            factor.to_numpy()[-period:],
            index=factor.index[:period].copy(),
            columns=factor.columns.copy(),
        )


def hold(
        factor: pd.DataFrame,
        *,
        period: int,
) -> pd.DataFrame:
    """Spread `factor` values for `period`.

    Can be used to react on new information every `period` timestamps.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    period : int
        Period to look back on the data.

    Returns
    -------
    pd.DataFrame
        Rolling mean of `factor` values.
    """

    periods = np.zeros(len(factor), dtype=int)
    update_periods = np.arange(len(factor), step=period)
    periods[update_periods] = update_periods
    update_mask = np.maximum.accumulate(periods[:, np.newaxis], axis=0)
    return pd.DataFrame(
        np.take_along_axis(np.asarray(factor, dtype=float), update_mask, axis=0),
        index=factor.index.copy(),
        columns=factor.columns.copy()
    )


def quantiles(
        factor: pd.DataFrame,
        *,
        min_q: float,
        max_q: float,
) -> pd.DataFrame:
    """Picks when `factor` values are between `min_q` and `max_q` quantiles in a period.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    min_q : float
        Quantile to estimate lower boarder of `factor` values to pick.
    max_q : float
        Quantile to estimate upper boarder of `factor` values to pick.

    Returns
    -------
    pd.DataFrame
        Matrix of True/False, indicating whether factor values are between quantile boarders or not.
    """

    factor_array = np.asarray(factor)
    lower, upper = np.nanquantile(factor_array, [min_q, max_q], axis=1, keepdims=True)
    return pd.DataFrame(
        (lower <= factor_array) & (factor_array <= upper),
        index=factor.index.copy(),
        columns=factor.columns.copy(),
    )


def top(
        factor: pd.DataFrame,
        *,
        k: int,
) -> pd.DataFrame:
    """Picks when `factor` values are at the top `k` in a period.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    k : int
        Place to estimate lower boarder of factor values to pick.

    Returns
    -------
    pd.DataFrame
        Matrix of True/False, indicating whether factor values are in the top or not.
    """

    def _top_k_row(arr: np.ndarray) -> float:
        uniq_arr = np.unique(arr[~np.isnan(arr)])
        max_k = len(uniq_arr)

        if max_k > k:
            return np.sort(uniq_arr)[-k]
        elif max_k > 0:
            return np.max(uniq_arr)
        else:
            return np.nan

    factor_array = np.asarray(factor)
    lower = np.apply_along_axis(_top_k_row, axis=1, arr=factor_array)[:, np.newaxis]
    return pd.DataFrame(
        factor_array >= lower,
        index=factor.index.copy(),
        columns=factor.columns.copy(),
    )


def bottom(
        factor: pd.DataFrame,
        *,
        k: int,
) -> pd.DataFrame:
    """Picks when `factor` values are in the bottom `k` in a period.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    k : int
        Place to estimate upper boarder of factor values to pick.

    Returns
    -------
    pd.DataFrame
        Matrix of True/False, indicating whether factor values are in the bottom or not.
    """

    def _bottom_k_row(arr: np.ndarray) -> float:
        uniq_arr = np.unique(arr[~np.isnan(arr)])
        max_k = len(uniq_arr)

        if max_k > k:
            return np.sort(uniq_arr)[k - 1]
        elif max_k > 0:
            return np.min(uniq_arr)
        else:
            return np.nan

    factor_array = np.asarray(factor)
    upper = np.apply_along_axis(_bottom_k_row, axis=1, arr=factor_array)[:, np.newaxis]
    return pd.DataFrame(
        factor_array <= upper,
        index=factor.index.copy(),
        columns=factor.columns.copy()
    )


def thresholds(
        factor: pd.DataFrame,
        *,
        min_t: float,
        max_t: float,
) -> pd.DataFrame:
    """Picks `factor` values between `min_t` and `max_t` thresholds.

    Parameters
    ----------
    factor : pd.DataFrame
        Matrix with factor values.
    min_t : float
        Lower boarder of `factor` values to pick.
    max_t : float
        Upper boarder of `factor` values to pick.

    Returns
    -------
    pd.DataFrame
        Matrix of True/False, indicating whether factor values are between thresholds or not.
    """

    factor_array = np.asarray(factor)
    return pd.DataFrame(
        (min_t <= factor_array) & (factor_array <= max_t),
        index=factor.index.copy(),
        columns=factor.columns.copy()
    )



"""Some additional stuff and "sugar" for creating pipelines."""

from __future__ import annotations

__all__ = [
    "align",
    "compose",
    "freeze",
]

from functools import reduce, partial as freeze
from typing import Callable, Any

import pandas as pd


def align(*args: pd.DataFrame | pd.Series) -> tuple[pd.DataFrame | pd.Series, ...]:
    """Aligns dataframes and series to ake them having the same index and columns.

    Parameters
    ----------
    args : sequence of pd.DataFrame or pd.Series
        Dataframes and series to be aligned.

    Returns
    -------
    tuple of pd.DataFrame or pd.Series
      Aligned dataframes and series.
    """

    args = list(args)
    for i in range(len(args) - 1):  # forward-aligning
        args[i], args[i + 1] = _align_two(args[i], args[i + 1])

    for i in range(len(args) - 2, 0, -1):  # backward-aligning
        args[i], args[i - 1] = _align_two(args[i], args[i - 1])
    return tuple(args)


def _are_aligned(
        x: pd.DataFrame | pd.Series,
        y: pd.DataFrame | pd.Series,
        /,
) -> tuple[pd.DataFrame | pd.Series, pd.DataFrame | pd.Series]:
    index_aligned = ((x.index.shape == y.index.shape) and (x.index == y.index).all())
    if isinstance(x, pd.Series) or isinstance(y, pd.Series):
        return index_aligned

    columns_aligned = ((x.columns.shape == y.columns.shape) and (x.columns == y.columns).all())
    return index_aligned and columns_aligned


def _align_two(
        x: pd.DataFrame | pd.Series,
        y: pd.DataFrame | pd.Series,
        /,
) -> tuple[pd.DataFrame | pd.Series, pd.DataFrame | pd.Series]:
    if _are_aligned(x, y):
        return x.copy(), y.copy()

    if isinstance(x, pd.Series) or isinstance(y, pd.Series):
        return x.align(y, join="inner", axis=0)

    return x.align(y, join="inner")


def compose(*steps) -> Callable[[Any, ...], Any]:
    """Combines functions to pipeline.

    Parameters
    ----------
    steps : sequence of callable
         Steps to be composed into pipeline.

    Returns
    -------
    callable
        Function, realizing full pipeline.
    """

    return reduce(lambda f, g: lambda *args, **kwargs: g(f(*args, **kwargs)), steps)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pqr']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.3,<2.0.0', 'pandas>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'pqr',
    'version': '1.0.0a0',
    'description': 'Lightweight library for backtesting factor strategies',
    'long_description': '# pqr\n\npqr is a python library for backtesting factor strategies. It is built in top of numpy, so it is \nfast and memory efficient, but provides pandas interface to make usage more convenient and verbose.\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install pqr.\n\n```bash\npip install pqr\n```\n\n## Quickstart\n\n```python\nimport pandas as pd\nimport pqr\n\nprices = pd.read_csv("prices.csv", parse_dates=True)\n\nmomentum = pqr.compose(\n    # picking\n    pqr.freeze(pqr.filter, universe=prices > 10),\n    pqr.freeze(pqr.look_back, period=12, agg="pct"),\n    pqr.freeze(pqr.lag, period=1),\n    pqr.freeze(pqr.hold, period=12),\n    pqr.freeze(pqr.quantiles, min_q=0.7, max_q=1),\n    # allocation\n    pqr.ew,\n    # evaluation\n    pqr.freeze(pqr.calculate, universe_returns=pqr.to_returns(prices)),\n)\n\n# returns series of returns of 30% ew momentum 12-1-12 strategy for stocks > 10$\nmomentum(prices)\n```\n\n## Documentation\n\nThe official documentation is hosted on readthedocs.org: https://pqr.readthedocs.io/en/latest/\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would \nlike to change.\n\nPlease make sure to update tests and documentation as appropriate.\n\n## License\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Andrey Babkin',
    'author_email': 'andrey.babkin.ru71@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pqr.readthedocs.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Basic operations to get weighted positions (holdings) from strategy signals."""

from __future__ import annotations

__all__ = [
    "allocate",
    "ew",
    "scale",
    "limit",
]

import numpy as np
import pandas as pd

from pqr.utils import align


def allocate(
        signals: pd.DataFrame,
        *,
        weights: pd.DataFrame,
) -> pd.DataFrame:
    """Calculates portfolio holdings.

    Weighted `signals` are normalized to 1.

    Parameters
    ----------
    signals : pd.DataFrame
        Matrix, consists of True/False, indicating presence of an asset in a portfolio.
    weights : pd.DataFrame
        Matrix with weights (e.g. market capitalization).

    Returns
    -------
    pd.DataFrame
        Matrix of holdings, each row sum equals to 1.
    """

    signals, weights = align(signals, weights)
    signals_array, weights_array = np.asarray(signals), np.asarray(weights)
    signals_array *= weights_array

    norm = np.nansum(signals_array, axis=1, keepdims=True, dtype=float)
    with np.errstate(divide="ignore", invalid="ignore"):
        return pd.DataFrame(
            np.nan_to_num(signals_array / norm, nan=0, neginf=0, posinf=0),
            index=signals.index.copy(),
            columns=signals.columns.copy(),
        )


def ew(signals: pd.DataFrame) -> pd.DataFrame:
    """Calculates equally-weighted holdings.

    Parameters
    ----------
    signals : pd.DataFrame
        Matrix, consists of True/False, indicating presence of an asset in a portfolio.

    Returns
    -------
    pd.DataFrame
        Matrix of holdings, each row sum equals to 1 and all non-zero row values are the same.
    """

    return allocate(signals, weights=signals)


def scale(
        holdings: pd.DataFrame,
        *,
        leverage: pd.Series,
) -> pd.DataFrame:
    """Calculates leveraged portfolio holdings.

    Parameters
    ----------
    holdings : pd.DataFrame
        Matrix of weighted positions.
    leverage : pd.DataFrame
        Series with leverage value in each period.

    Returns
    -------
    pd.DataFrame
        Leveraged `holdings`.
    """

    holdings, leverage = align(holdings, leverage)
    return pd.DataFrame(
        np.asarray(holdings) * np.asarray(leverage)[:, np.newaxis],
        index=holdings.index.copy(),
        columns=holdings.columns.copy(),
    )


def limit(
        holdings: pd.DataFrame,
        *,
        min_leverage: float,
        max_leverage: float,
) -> pd.DataFrame:
    """Clips portfolio leverage by min and max allowed leverage in a period.

    Parameters
    ----------
    holdings : pd.DataFrame
        Matrix of weighted positions.
    min_leverage : float
        Minimum allowed total leverage in a period.
    max_leverage : float
        Maximum allowed total leverage in a period.

    Returns
    -------
    pd.DataFrame
        Matrix with scaled weights.
    """

    total_leverage = np.nansum(holdings, axis=1, keepdims=True)
    too_low = total_leverage < min_leverage
    too_high = total_leverage > max_leverage

    corrector = np.ones_like(total_leverage, dtype=float)
    with np.errstate(divide="ignore", invalid="ignore"):
        corrector[too_low] *= min_leverage / total_leverage[too_low]
        corrector[too_high] *= max_leverage / total_leverage[too_high]
    corrector = np.nan_to_num(corrector, nan=1, neginf=1, posinf=1)

    corrector = pd.Series(corrector, index=holdings.index.copy())
    return scale(holdings, leverage=corrector)
